from django import forms
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError

from content_access_control.model_identifiers import str_to_instance, instance_to_str
from content_access_control.models import (
    ContentAccessPermission,
    PolicySubject,
    PolicySubjectGroup,
)


class _ResourcePermissionAdminFactory:
    def __init__(
        self,
        resource_model,
        actions: list[str],
        app_label: str | None = None,
        **kwargs,
    ):
        self.subject_models = [
            (PolicySubject, "PolicySubject"),
            (PolicySubjectGroup, "SubjectGroup"),
        ]
        self.resource_model = resource_model
        self.resource_label = resource_model.__name__
        self.actions = actions
        self.app_label = app_label or resource_model._meta.app_label
        self.subject_widget = kwargs.get("subject_widget", None)
        self.resource_widget = kwargs.get("resource_widget", None)

        # define the raw form fields
        self.form_fields = {
            "subject_obj": forms.ChoiceField(
                label="Subject",
                required=True,
                widget=self.subject_widget,
            ),
            "resource_obj": forms.ChoiceField(
                label="Resource",
                required=True,
                widget=self.resource_widget,
            ),
        }
        if len(self.actions) > 1:
            self.form_fields["action"] = forms.ChoiceField(
                label="Action",
                choices=[(a, a.capitalize()) for a in self.actions],
                required=True,
            )

    def __call__(self):
        return (
            self._build_content_access_permission_proxy(),
            self._build_admin_model_for_model_proxy(),
        )

    def _build_content_access_permission_proxy(self):
        """
        Dynamically creates a proxy of the permission_model.
        """

        class Meta:
            proxy = True
            app_label = self.app_label

        # make a name by concatenating the subject class names
        name = f"{self.resource_label}ContentAccessPermission"

        attrs = {
            "__module__": self.resource_model.__module__,
            "Meta": Meta,
            "__doc__": f"Proxy for {self.resource_model.__name__} access permision model in {self.app_label}",
        }
        return type(name, (ContentAccessPermission,), attrs)

    def _build_admin_form_for_model_proxy(self):
        """
        Dynamically builds a ModelForm tied to the proxy model.
        """
        proxy_model = self._build_content_access_permission_proxy()
        class_name = proxy_model.__name__ + "AdminForm"

        # Capture into locals to avoid late binding
        subject_models = self.subject_models
        resource_model = self.resource_model
        resource_label = self.resource_label
        actions = self.actions

        # Copy the base fields
        form_fields = dict(self.form_fields)

        # __init__ will populate choices and handle initial/action defaults
        def __init__(form_self, *args, **kwargs):
            forms.ModelForm.__init__(form_self, *args, **kwargs)

            # If using the default widget, populate choices. Otherwise, assume AJAX or custom choices.
            if self.subject_widget is None:
                subj_choices = []
                for model, prefix in subject_models:
                    subj_choices += [
                        (instance_to_str(obj), f"{prefix}: {obj}")
                        for obj in model.objects.all()
                    ]
                form_self.fields["subject_obj"].choices = subj_choices

            if self.resource_widget is None:
                res_choices = [
                    (instance_to_str(obj), f"{resource_label}: {obj}")
                    for obj in resource_model.objects.all()
                ]
                form_self.fields["resource_obj"].choices = res_choices

            # if editing an existing instance, set initial values
            if form_self.instance.pk:
                form_self.fields["subject_obj"].initial = instance_to_str(
                    form_self.instance.subject
                )
                form_self.fields["resource_obj"].initial = instance_to_str(
                    form_self.instance.resource
                )
                if len(actions) > 1:
                    form_self.fields["action"].initial = getattr(
                        form_self.instance, "action", actions[0]
                    )

            # if exactly one action, force it onto the instance
            if len(actions) == 1:
                form_self.instance.action = actions[0]
            elif len(actions) == 0:
                form_self.instance.action = ""

        def clean(form_self):
            chosen_sub = str_to_instance(form_self.cleaned_data["subject_obj"])
            sub_ct = ContentType.objects.get_for_model(type(chosen_sub))
            form_self.instance.subject_content_type = sub_ct
            form_self.instance.subject_id = chosen_sub.pk

            chosen_res = str_to_instance(form_self.cleaned_data["resource_obj"])
            res_ct = ContentType.objects.get_for_model(type(chosen_res))
            form_self.instance.resource_content_type = res_ct
            form_self.instance.resource_id = chosen_res.pk

            # if there are multiple actions, pick up the selected one
            if len(actions) > 1:
                form_self.instance.action = form_self.cleaned_data["action"]
            elif len(actions) == 1:
                form_self.instance.action = actions[0]
            else:
                form_self.instance.action = ""

            cleaned_data = forms.ModelForm.clean(form_self)

            if (
                ContentAccessPermission.objects.filter(
                    subject_content_type=sub_ct,
                    subject_id=chosen_sub.pk,
                    resource_content_type=res_ct,
                    resource_id=chosen_res.pk,
                    action=form_self.instance.action,
                )
                .exclude(pk=form_self.instance.pk)
                .exists()
            ):
                raise ValidationError("This Access Permission already exists.")

            return cleaned_data

        # build Meta inner class
        Meta = type(
            "Meta",
            (),
            {
                "model": proxy_model,
                "fields": list(form_fields.keys()),
            },
        )

        # assemble all attrs into the dynamic form class
        attrs = {
            **form_fields,
            "Meta": Meta,
            "__init__": __init__,
            "clean": clean,
            "__module__": self.resource_model.__module__,
        }

        return type(class_name, (forms.ModelForm,), attrs)

    def _build_admin_model_for_model_proxy(self):
        class_name = f"{self.resource_label}ContentAccessPermissionAdmin"

        def get_queryset(_self, request):
            ct = ContentType.objects.get_for_model(self.resource_model)
            qs = admin.ModelAdmin.get_queryset(_self, request)
            return qs.filter(resource_content_type=ct)

        attrs = {
            "form": self._build_admin_form_for_model_proxy(),
            "get_queryset": get_queryset,
        }

        return type(class_name, (admin.ModelAdmin,), attrs)


def register_permission_admin(model, actions, **kwargs):
    """
    Creates and registers a dynamic ModelAdmin for managing
    content access permissions for the given model.
    """
    factory = _ResourcePermissionAdminFactory(model, actions, **kwargs)
    proxy_model, admin_model = factory()
    admin.site.register(proxy_model, admin_model)



import logging
from django.conf import settings

from casbin import Enforcer

from .utils import import_class

logger = logging.getLogger(__name__)


class ProxyEnforcer(Enforcer):
    _initialized = False
    db_alias = "default"

    def __init__(self, *args, **kwargs):
        if self._initialized:
            super().__init__(*args, **kwargs)
        else:
            logger.info(
                "Deferring casbin enforcer initialisation until django is ready"
            )

    def _load(self):
        if self._initialized is False:
            logger.info("Performing deferred casbin enforcer initialisation")
            self._initialized = True
            model = str(getattr(settings, "CASBIN_MODEL"))
            adapter_loc = getattr(
                settings, "CASBIN_ADAPTER", "content_access_control.adapter.Adapter"
            )
            adapter_args = getattr(settings, "CASBIN_ADAPTER_ARGS", tuple())
            self.db_alias = getattr(settings, "CASBIN_DB_ALIAS", "default")
            Adapter = import_class(adapter_loc)
            adapter = Adapter(self.db_alias, *adapter_args)

            super().__init__(model, adapter)
            logger.debug("Casbin enforcer initialised")

            watcher = getattr(settings, "CASBIN_WATCHER", None)
            if watcher:
                self.set_watcher(watcher)

            role_manager = getattr(settings, "CASBIN_ROLE_MANAGER", None)
            if role_manager:
                self.set_role_manager(role_manager)

    def __getattribute__(self, name):
        safe_methods = ["__init__", "_load", "_initialized", "db_alias"]
        if not super().__getattribute__("_initialized") and name not in safe_methods:
            self._load()

        return super().__getattribute__(name)



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

from .core import enforcer

UserModel = get_user_model()


class CasbinBackend(ModelBackend):
    """
    Check permissions with Casbin.
    """

    def __init__(self):
        self.enforcer = enforcer

    def user_can_authenticate(self, user):
        """
        Reject users with is_active=False. Custom user models that don't have
        that attribute are allowed.
        """
        return getattr(user, "is_active", True)

    def _get_permissions(self, user_obj, obj, from_name):
        """
        Return the direct permissions of `user_obj`
        """
        if not user_obj.is_active or user_obj.is_anonymous or obj is not None:
            return set()

        perm_cache_name = f"_{from_name}_perm_cache"
        if not hasattr(user_obj, perm_cache_name):
            policies = self.enforcer.get_implicit_permissions_for_user(
                user_obj.username
            )
            perms = tuple(map(tuple, policies))
            setattr(user_obj, perm_cache_name, perms)
        return getattr(user_obj, perm_cache_name)

    def get_user_permissions(self, user_obj, obj=None):
        """
        Return a set of permission the user `user_obj` has from their
        `user_permissions`.
        """
        policies = self.enforcer.get_permissions_for_user(user_obj.username)
        return tuple(map(tuple, policies))

    def get_all_permissions(self, user_obj, obj=None):
        """
        Return a set of permission the user `user_obj` and inherited roles have.
        The result is cached for each user. Refresh By requesting a new instance.
        """
        if not user_obj.is_active or user_obj.is_anonymous or obj is not None:
            return set()
        return self._get_permissions(user_obj, obj, from_name="user")

    def has_perm(self, user_obj, perm, obj=None):
        return user_obj.is_active and super().has_perm(user_obj, perm, obj=obj)

    def get_user(self, user_id):
        try:
            user = UserModel._default_manager.get(pk=user_id)
        except UserModel.DoesNotExist:
            return None
        return user if self.user_can_authenticate(user) else None



from django.contrib import admin

from .models import (
    CasbinRule,
    Feature,
    PolicySubject,
    PolicySubjectGroup,
    SubjectToGroup,
)
from .admin_permission import register_permission_admin


class SubjectToGroupInline(admin.TabularInline):
    model = SubjectToGroup
    extra = 0


@admin.register(CasbinRule)
class CasbinRuleAdmin(admin.ModelAdmin): ...


@admin.register(PolicySubject)
class PolicySubjectAdmin(admin.ModelAdmin):
    inlines = [SubjectToGroupInline]


@admin.register(PolicySubjectGroup)
class PolicySubjectGroupAdmin(admin.ModelAdmin):
    inlines = [SubjectToGroupInline]


@admin.register(Feature)
class FeatureAdmin(admin.ModelAdmin): ...


register_permission_admin(Feature, ["access"])



from django.apps import AppConfig
from django.core.exceptions import ImproperlyConfigured


class ContentAccessControlConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "content_access_control"

    def ready(self):
        from django.conf import settings  # noqa
        from .core import enforcer  # noqa

        try:
            getattr(settings, "CASBIN_MODEL")
        except AttributeError:
            raise ImproperlyConfigured("CASBIN_MODEL must be set in settings.py")

        enforcer._load()



import logging

from rest_framework.permissions import BasePermission
from .core import enforcer


logger = logging.getLogger(__name__)


class SubjectHasUrlPermission(BasePermission):
    """
    Custom DRF permission class using dauthz to enforce URL-based access control.
    """

    def has_permission(self, request, view):
        if hasattr(request, "policy_subject"):
            subject = request.policy_subject.unique_object_instance_identifier
        else:
            if hasattr(request, "user"):
                subject = str(request.user.username)
            else:
                subject = "anonymous"
        resource = request.path
        action = request.method
        can_access = enforcer.enforce(subject, resource, action)
        logger.debug(
            f"Verified access for {resource}, and {action=}, and {subject=} --> Result: {can_access}"
        )
        return can_access



import logging

from casbin import persist
from django.db.utils import OperationalError, ProgrammingError

from .models import CasbinRule

logger = logging.getLogger(__name__)


class Adapter(persist.Adapter):
    """the interface for Casbin adapters."""

    def __init__(self, db_alias="default"):
        self.db_alias = db_alias

    def load_policy(self, model):
        """loads all policy rules from the storage."""
        try:
            lines = CasbinRule.objects.using(self.db_alias).all()

            for line in lines:
                persist.load_policy_line(str(line), model)
        except (OperationalError, ProgrammingError) as error:
            logger.warning("Could not load policy from database: {}".format(error))

    def _create_policy_line(self, ptype, rule):
        line = CasbinRule(ptype=ptype)
        if len(rule) > 0:
            line.v0 = rule[0]
        if len(rule) > 1:
            line.v1 = rule[1]
        if len(rule) > 2:
            line.v2 = rule[2]
        if len(rule) > 3:
            line.v3 = rule[3]
        if len(rule) > 4:
            line.v4 = rule[4]
        if len(rule) > 5:
            line.v5 = rule[5]
        return line

    def save_policy(self, model):
        """saves all policy rules to the storage."""
        # See https://casbin.org/docs/adapters/#autosave
        # for why this is deleting all rules
        CasbinRule.objects.using(self.db_alias).all().delete()

        lines = []
        for sec in ["p", "g"]:
            if sec not in model.model.keys():
                continue
            for ptype, ast in model.model[sec].items():
                for rule in ast.policy:
                    lines.append(self._create_policy_line(ptype, rule))
        rows_created = CasbinRule.objects.using(self.db_alias).bulk_create(lines)
        return len(rows_created) > 0

    def add_policy(self, sec, ptype, rule):
        """adds a policy rule to the storage."""
        line = self._create_policy_line(ptype, rule)
        line.save()

    def remove_policy(self, sec, ptype, rule):
        """removes a policy rule from the storage."""
        query_params = {"ptype": ptype}
        for i, v in enumerate(rule):
            query_params["v{}".format(i)] = v
        rows_deleted, _ = (
            CasbinRule.objects.using(self.db_alias).filter(**query_params).delete()
        )
        return rows_deleted > 0

    def remove_filtered_policy(self, sec, ptype, field_index, *field_values):
        """removes policy rules that match the filter from the storage.
        This is part of the Auto-Save feature.
        """
        query_params = {"ptype": ptype}
        if not (0 <= field_index <= 5):
            return False
        if not (1 <= field_index + len(field_values) <= 6):
            return False
        for i, v in enumerate(field_values):
            query_params["v{}".format(i + field_index)] = v
        rows_deleted, _ = (
            CasbinRule.objects.using(self.db_alias).filter(**query_params).delete()
        )
        return rows_deleted > 0



from django.apps import apps
from django.db import models


def instance_to_str(instance: models.Model) -> str:
    """
    Convert a model instance into a unique string of the form:
    "app_label:model_name:pk"
    """
    assert instance is not None, (
        "This function cannot accept NoneType, pass models.Model instead"
    )
    app_label = instance._meta.app_label
    model_name = instance._meta.model_name
    pk = instance.pk
    return f"{app_label}:{model_name}:{pk}"


def str_to_instance(instance_str: str) -> models.Model:
    """
    Given a string of the form "app_label:model_name:pk", return the corresponding model instance.
    Raises DoesNotExist if the object does not exist.
    """
    try:
        app_label, model_name, pk_str = instance_str.split(":")
    except ValueError:
        raise ValueError("String format must be 'app_label:model_name:pk'")

    model = apps.get_model(app_label=app_label, model_name=model_name)
    return model.objects.get(pk=pk_str)



# Generated by Django 5.2.6 on 2025-09-09 20:29

import content_access_control.policy_mixins
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="CasbinRule",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("ptype", models.CharField(max_length=255)),
                ("v0", models.CharField(blank=True, max_length=255)),
                ("v1", models.CharField(blank=True, max_length=255)),
                ("v2", models.CharField(blank=True, max_length=255)),
                ("v3", models.CharField(blank=True, max_length=255)),
                ("v4", models.CharField(blank=True, max_length=255)),
                ("v5", models.CharField(blank=True, max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="ContentAccessPermission",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("subject_id", models.PositiveIntegerField()),
                ("resource_id", models.PositiveIntegerField()),
                ("action", models.CharField(max_length=255)),
                (
                    "resource_content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="resource_type",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "subject_content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="subject_type",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            bases=(
                content_access_control.policy_mixins.AddRemovePermissionPolicyMixin,
                content_access_control.policy_mixins.AddRemovePolicyMixin,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="FeatureContentAccessPermission",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("content_access_control.contentaccesspermission",),
        ),
        migrations.CreateModel(
            name="Feature",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(
                        fields=("name",), name="unique_feature_name_constraint"
                    )
                ],
            },
            bases=(
                content_access_control.policy_mixins.ObjectIdentifierMixin,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="PolicySubject",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(default="default", max_length=255)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                content_access_control.policy_mixins.AddRemoveGroupingPolicyMixin,
                content_access_control.policy_mixins.AddRemovePolicyMixin,
                content_access_control.policy_mixins.ObjectIdentifierMixin,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="PolicySubjectGroup",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "parent_group",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="content_access_control.policysubjectgroup",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                content_access_control.policy_mixins.AddRemoveGroupingPolicyMixin,
                content_access_control.policy_mixins.ObjectIdentifierMixin,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="SubjectToGroup",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "subject",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="content_access_control.policysubject",
                    ),
                ),
                (
                    "subject_group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="content_access_control.policysubjectgroup",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                content_access_control.policy_mixins.AddRemoveGroupingPolicyMixin,
                content_access_control.policy_mixins.AddRemovePolicyMixin,
                content_access_control.policy_mixins.ObjectIdentifierMixin,
                models.Model,
            ),
        ),
        migrations.AddConstraint(
            model_name="contentaccesspermission",
            constraint=models.UniqueConstraint(
                fields=(
                    "subject_content_type",
                    "subject_id",
                    "resource_content_type",
                    "resource_id",
                    "action",
                ),
                name="unique_subject_resource",
            ),
        ),
    ]



import importlib


def import_class(name):
    """Import class from string
    e.g. `package.module.ClassToImport` returns the `ClassToImport` class"""
    components = name.split(".")
    module_name = ".".join(components[:-1])
    class_name = components[-1]
    module = importlib.import_module(module_name)
    return getattr(module, class_name)



from django.db import models
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey

from content_access_control.policy_mixins import (
    AddRemovePermissionPolicyMixin,
    AddRemoveGroupingPolicyMixin,
    PolicyLifecycleMixin,
    ObjectIdentifierMixin,
    SubjectAccessPermissionMixin,
    ResourceAccessPermissionMixin,
)


class ExpectedModel(ObjectIdentifierMixin, models.Model):
    class Meta:
        abstract = True


class CasbinRule(models.Model):
    ptype = models.CharField(max_length=255)
    v0 = models.CharField(max_length=255, blank=True)
    v1 = models.CharField(max_length=255, blank=True)
    v2 = models.CharField(max_length=255, blank=True)
    v3 = models.CharField(max_length=255, blank=True)
    v4 = models.CharField(max_length=255, blank=True)
    v5 = models.CharField(max_length=255, blank=True)

    def __str__(self):
        text = self.ptype

        if self.v0:
            text = text + ", " + self.v0
        if self.v1:
            text = text + ", " + self.v1
        if self.v2:
            text = text + ", " + self.v2
        if self.v3:
            text = text + ", " + self.v3
        if self.v4:
            text = text + ", " + self.v4
        if self.v5:
            text = text + ", " + self.v5
        return text

    def __repr__(self):
        return '<CasbinRule {}: "{}">'.format(self.id, str(self))


class ContentAccessPermission(
    PolicyLifecycleMixin, AddRemovePermissionPolicyMixin, models.Model
):
    """
    Model used to encapsulate a generic link between
    "which *subject* can do which *action* on which *resource*"
    """

    subject_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="subject_type"
    )
    subject_id = models.PositiveIntegerField()
    subject = GenericForeignKey("subject_content_type", "subject_id")
    resource_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="resource_type"
    )
    resource_id = models.PositiveIntegerField()
    resource = GenericForeignKey("resource_content_type", "resource_id")
    action = models.CharField(max_length=255)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=[
                    "subject_content_type",
                    "subject_id",
                    "resource_content_type",
                    "resource_id",
                    "action",
                ],
                name="unique_subject_resource",
            )
        ]

    @property
    def access_policy_identifiers(self):
        return (
            get_uuid_of_model(self.subject),
            get_uuid_of_model(self.resource),
            self.action,
        )

    def __str__(self):
        # Use ContentType to avoid accessing the actual related objects
        try:
            sub_app_label = self.subject_content_type.app_label
            sub_model = self.subject_content_type.model
            sub_label = f"{sub_app_label}.{sub_model}"
        except Exception:
            sub_label = "unknown.subject"

        try:
            res_app_label = self.resource_content_type.app_label
            res_model = self.resource_content_type.model
            res_label = f"{res_app_label}.{res_model}"
        except Exception:
            res_label = "unknown.resource"

        # Prefer IDs to avoid fetching GenericForeignKey instances.
        # If you still want best-effort human-readable names without hard failures,
        # you can try to use str(self.subject) / str(self.resource) guarded by try/except.
        return f'{sub_label}({self.subject_id}) can do: "{self.action}" on {res_label}({self.resource_id})'


class PolicySubject(
    SubjectAccessPermissionMixin,
    PolicyLifecycleMixin,
    AddRemoveGroupingPolicyMixin,
    ObjectIdentifierMixin,
    models.Model,
):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    name = models.CharField(max_length=255, default="default")

    @property
    def access_policy_identifiers(self):
        return (self.user.username, self.unique_object_instance_identifier)

    def __str__(self):
        return f"{self.user.username} - {self.name}"


class PolicySubjectGroup(
    SubjectAccessPermissionMixin,
    AddRemoveGroupingPolicyMixin,
    ObjectIdentifierMixin,
    models.Model,
):
    name = models.CharField(max_length=255)
    parent_group = models.ForeignKey(
        "self", on_delete=models.CASCADE, null=True, blank=True
    )

    @property
    def access_policy_identifiers(self):
        return (
            self.unique_object_instance_identifier,
            self.parent_group.unique_object_instance_identifier,
        )

    def __str__(self):
        parent = str(self.parent_group or "")
        SEP = " - " if parent else ""
        return f"{parent}{SEP}{self.name}"

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        if not self.parent_group:
            return
        self.remove_access_permission_policy()

    def save(self, *args, **kwargs):
        if self.pk:
            # Check if not self-referential
            if self.parent_network_id == self.pk:
                raise ValueError(
                    "Network parent cannot reference the network itself, choose different parent or set to null"
                )

            # Delete the old grouping policy
            old_instance = PolicySubjectGroup.objects.get(pk=self.pk)
            if old_instance.parent_network:
                old_instance.remove_access_permission_policy()

        super().save(*args, **kwargs)
        if not self.parent_group:
            return
        self.add_access_permission_policy()


class SubjectToGroup(
    PolicyLifecycleMixin,
    AddRemoveGroupingPolicyMixin,
    ObjectIdentifierMixin,
    models.Model,
):
    subject = models.ForeignKey(PolicySubject, on_delete=models.CASCADE)
    subject_group = models.ForeignKey(PolicySubjectGroup, on_delete=models.CASCADE)

    @property
    def access_policy_identifiers(self):
        return (
            self.subject.unique_object_instance_identifier,
            self.subject_group.unique_object_instance_identifier,
        )

    def __str__(self):
        return f'Subject: "{self.subject}", assigned to: "{self.subject_group}"'


ModelType = models.Model | ExpectedModel


def get_uuid_of_model(model_instance: ModelType):
    if hasattr(model_instance, "unique_object_instance_identifier"):
        return model_instance.unique_object_instance_identifier
    return str(model_instance)


class Feature(
    ResourceAccessPermissionMixin,
    ObjectIdentifierMixin,
    models.Model,
):
    name = models.CharField(max_length=255)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["name"], name="unique_feature_name_constraint"
            )
        ]

    def __str__(self):
        return self.name



import logging
from django.db import models
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.contenttypes.fields import GenericRelation

from .model_identifiers import instance_to_str
from .core import enforcer


logger = logging.getLogger(__name__)


class AddRemovePolicyMixin:
    """
    Mixin for adding and removing policies.

    Do not use this mixin directly, unless you know what you're doing.
    More likely than not, you should use one of the subclasses:
    AddRemovePermissionPolicyMixin or AddRemoveGroupingPolicyMixin.
    This is because the mixin does not define the function for adding and removing policies.
    """

    @property
    def policy_add_function(self):
        raise NotImplementedError("This needs to be implemented by each class")

    @property
    def policy_remove_function(self):
        raise NotImplementedError("This needs to be implemented by each class")

    @property
    def access_policy_identifiers(self):
        raise NotImplementedError("This needs to be implemented by each class")

    def add_access_permission_policy(self) -> bool:
        identifiers = self.access_policy_identifiers
        logger.debug(f"Adding a policy with {identifiers=}")
        return self.policy_add_function(*identifiers)

    def remove_access_permission_policy(self) -> bool:
        identifiers = self.access_policy_identifiers
        logger.debug(f"Removing a policy with {identifiers=}")
        return self.policy_remove_function(*identifiers)


class AddRemovePermissionPolicyMixin(AddRemovePolicyMixin):
    @property
    def policy_add_function(self):
        return enforcer.add_policy

    @property
    def policy_remove_function(self):
        return enforcer.remove_policy


class AddRemoveGroupingPolicyMixin(AddRemovePolicyMixin):
    @property
    def policy_add_function(self):
        return enforcer.add_grouping_policy

    @property
    def policy_remove_function(self):
        return enforcer.remove_grouping_policy


class PolicyLifecycleMixin(AddRemovePolicyMixin, models.Model):
    """
    Mixin that:
      - On create: adds policy after save.
      - On update: removes policy tied to the old instance, then adds policy for the new state.
      - On delete: removes policy, then deletes the row.

    Requirement on the concrete model is that it must inherit from either
    AddRemovePermissionPolicyMixin or AddRemoveGroupingPolicyMixin.
    """

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        old_instance = None
        is_update = self.pk is not None

        if is_update:
            try:
                # Get the old persisted state to remove its policy
                old_instance = self.__class__.objects.get(pk=self.pk)
            except ObjectDoesNotExist:
                logger.warning(
                    f"Object with {self.__class__.__name__=} and {self.pk=} had .save() method called on it, but it does not exist in the database."
                )
                """
                There has been an error, but it makes no sense, so we continue,
                most of the time it should not happen and the rest of the time
                it should not be an issue.
                """

        if old_instance is not None:
            # Remove policy based on the old state
            old_instance.remove_access_permission_policy()

        # Persist the new state
        super().save(*args, **kwargs)

        # Add policy based on the new state
        self.add_access_permission_policy()

    def delete(self, *args, **kwargs):
        # Remove policy based on current state before deleting
        self.remove_access_permission_policy()
        super().delete(*args, **kwargs)


class ObjectIdentifierMixin:
    @property
    def unique_object_instance_identifier(self) -> str:
        return instance_to_str(self)


class ResourceAccessPermissionMixin(models.Model):
    # standardize the relation name to `content_access_permission`
    content_access_permission = GenericRelation(
        "content_access_control.ContentAccessPermission",
        content_type_field="resource_content_type",
        object_id_field="resource_id",
        related_query_name="resource",
    )

    class Meta:
        abstract = True


class SubjectAccessPermissionMixin(models.Model):
    content_access_permission = GenericRelation(
        "content_access_control.ContentAccessPermission",
        content_type_field="subject_content_type",
        object_id_field="subject_id",
        related_query_name="subject",
    )

    class Meta:
        abstract = True



from content_access_control.models import PolicySubject


class PolicySubjectMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        subject_name = request.headers.get("X-Policy-Subject-Act-As", "default")
        policy_subject = None

        # Check if user is authenticated
        if request.user.is_authenticated:
            try:
                policy_subject = PolicySubject.objects.get(
                    user=request.user, name=subject_name
                )
            except PolicySubject.DoesNotExist:
                policy_subject = None

        request.policy_subject = policy_subject
        return self.get_response(request)
